"""
Core Types for the Library
"""

import os
import inspect
import pathlib
import tempfile
import atexit

from fileio.lib.core import pathlib, IterableAIOFile
from fileio.lib.base import *
from fileio.lib.posix.cloud import CloudFileSystemPath
from fileio.providers.gcs import *
from fileio.providers.s3 import *
from fileio.providers.minio import *
from fileio.providers.s3c import *
from fileio.providers.r2 import *
from fileio.providers.wasabi import *
from fileio.providers.azure import *

from fileio.lib.apis import (
    FastUploadFile,
    StarletteUploadFile,
    StarliteUploadFile,
)
from fileio.types.classprops import lazyproperty
from fileio.types.options import LoadMode
from fileio.utils.logs import logger
from fileio.utils.pooler import ThreadPooler
from typing import Union, Any, TypeVar, List, Optional, Callable, Dict, Type, Tuple, TYPE_CHECKING

PathLikeT = Union[
    FilePurePath,
    FilePath,
    PureFilePosixPath,
    FileWindowsPath,
    FilePosixPath,
    PureFileWindowsPath,
    
    Type[FilePurePath],
    Type[FilePath],
    Type[PureFilePosixPath],
    Type[FileWindowsPath],
    Type[FilePosixPath],
    Type[PureFileWindowsPath]
]

GSPathLikeT = Union[
    Type[FileGSPurePath],
    Type[FileGSPath],
    Type[PureFileGSPosixPath],
    Type[FileGSWindowsPath],
    Type[FileGSPosixPath],
    Type[PureFileGSWindowsPath],
]

AzurePathLikeT = Union[
    Type[FileAZPurePath],
    Type[FileAZPath],
    Type[PureFileAZPosixPath],
    Type[FileAZWindowsPath],
    Type[FileAZPosixPath],
    Type[PureFileAZWindowsPath],
]

S3PathLikeT = Union[
    Type[FileS3PurePath],
    Type[FileS3Path],
    Type[PureFileS3PosixPath],
    Type[FileS3WindowsPath],
    Type[FileS3PosixPath],
    Type[PureFileS3WindowsPath],
]

MinioPathLikeT = Union[
    Type[FileMinioPurePath], 
    Type[FileMinioPath], 
    Type[PureFileMinioPosixPath], 
    Type[FileMinioWindowsPath], 
    Type[FileMinioPosixPath], 
    Type[PureFileMinioWindowsPath],
]

S3CPathLikeT = Union[
    Type[FileS3CPurePath],
    Type[FileS3CPath],
    Type[PureFileS3CPosixPath],
    Type[FileS3CWindowsPath],
    Type[FileS3CPosixPath],
    Type[PureFileS3CWindowsPath],
]


R2PathLikeT = Union[
    Type[FileR2PurePath],
    Type[FileR2Path],
    Type[PureFileR2PosixPath],
    Type[FileR2WindowsPath],
    Type[FileR2PosixPath],
    Type[PureFileR2WindowsPath],
]

WasabiPathLikeT = Union[
    Type[FileWasabiPurePath],
    Type[FileWasabiPath],
    Type[PureFileWasabiPosixPath],
    Type[FileWasabiWindowsPath],
    Type[FileWasabiPosixPath],
    Type[PureFileWasabiWindowsPath],
]


# FileLike = Union[
#     Type[FilePurePath],
#     Type[FilePath],
#     Type[PureFilePosixPath],
#     Type[FileWindowsPath],
#     Type[FilePosixPath],
#     Type[PureFileWindowsPath],
    
#     Type[FileGSPurePath],
#     Type[FileGSPath],
#     Type[PureFileGSPosixPath],
#     Type[FileGSWindowsPath],
#     Type[FileGSPosixPath],
#     Type[PureFileGSWindowsPath],
    
#     Type[FileS3PurePath],
#     Type[FileS3Path],
#     Type[PureFileS3PosixPath],
#     Type[FileS3WindowsPath],
#     Type[FileS3PosixPath],
#     Type[PureFileS3WindowsPath],

#     Type[FileMinioPurePath], 
#     Type[FileMinioPath], 
#     Type[PureFileMinioPosixPath], 
#     Type[FileMinioWindowsPath], 
#     Type[FileMinioPosixPath], 
#     Type[PureFileMinioWindowsPath],

#     Type[FileS3CPurePath],
#     Type[FileS3CPath],
#     Type[PureFileS3CPosixPath],
#     Type[FileS3CWindowsPath],
#     Type[FileS3CPosixPath],
#     Type[PureFileS3CWindowsPath],

# ]
FileLike = PathLikeT
FileLikeT = TypeVar("FileLikeT", bound=FileLike)
# FileLike = Union[
#     PathLikeT,
#     GSPathLikeT,
#     S3PathLikeT,
#     MinioPathLikeT,
#     S3CPathLikeT,
#     R2PathLikeT,
#     WasabiPathLikeT,
# ]

_PATHLIKE_CLS: Tuple[FileLike, ...] = (
    FilePurePath,
    FilePath,
    PureFilePosixPath,
    FileWindowsPath,
    FilePosixPath,
    PureFileWindowsPath,
    
    FileGSPurePath,
    FileGSPath,
    PureFileGSPosixPath,
    FileGSWindowsPath,
    FileGSPosixPath,
    PureFileGSWindowsPath,
    
    FileS3PurePath,
    FileS3Path,
    PureFileS3PosixPath,
    FileS3WindowsPath,
    FileS3PosixPath,
    PureFileS3WindowsPath,

    FileAZPurePath,
    FileAZPath,
    PureFileAZPosixPath,
    FileAZWindowsPath,
    FileAZPosixPath,
    PureFileAZWindowsPath,

    FileMinioPurePath, 
    FileMinioPath, 
    PureFileMinioPosixPath, 
    FileMinioWindowsPath, 
    FileMinioPosixPath, 
    PureFileMinioWindowsPath,

    FileS3CPurePath,
    FileS3CPath,
    PureFileS3CPosixPath,
    FileS3CWindowsPath,
    FileS3CPosixPath,
    PureFileS3CWindowsPath,

    FileR2PurePath,
    FileR2Path,
    PureFileR2PosixPath,
    FileR2WindowsPath,
    FileR2PosixPath,

    FileWasabiPurePath,
    FileWasabiPath,
    PureFileWasabiPosixPath,
    FileWasabiWindowsPath,
    FileWasabiPosixPath,
)

FileSysLike = Union[
    Type[AWSFileSystem],
    Type[GCPFileSystem],
    Type[AZFileSystem],
    Type[MinioFileSystem],
    Type[S3CFileSystem],
    Type[R2FileSystem],
    Type[WasabiFileSystem],

]

PathLike = Union[str, os.PathLike, FileLike]

FileType = TypeVar(
    'FileType', 
    str, 
    pathlib.Path, 
    FileLike, 
    FastUploadFile, 
    StarletteUploadFile, 
    StarliteUploadFile,
    Any
)

FileListType = TypeVar(
    'FileListType', 
    List[
        Union[
            str, 
            pathlib.Path, 
            FileLike, 
            FastUploadFile, 
            StarletteUploadFile, 
            StarliteUploadFile, 
            Any
        ]
    ],
    List[str], 
    List[pathlib.Path], 
    List[FileLike], 
    List[FastUploadFile], 
    List[StarletteUploadFile], 
    List[StarliteUploadFile],
    List[Any]
)


_PREFIXES_TO_CLS: Dict[str, FileLike] = {
    'gs://': FileGSPath,
    's3://': FileS3Path,
    'az://': FileAZPath,
    #'s3a://': FileMinioPath,
    'mio://': FileMinioPath,
    'minio://': FileMinioPath,

    's3compat://': FileS3CPath,
    's3c://': FileS3CPath,
    #'minio://': cloud.PosixMinioPath,
    #'s3compat://': cloud.PosixS3CompatPath,
    'r2://': FileR2Path,
    'wsbi://': FileWasabiPath,
}


def as_path(path: PathLike) -> FileLike:
    """
    Given a path-like object, return a path-like object
    
    Create a generic `pathlib.Path`-like abstraction.
    Depending on the input (e.g. `gs://`, `github://`, `ResourcePath`,...), the
    system (Windows, Linux,...), the function will create the right pathlib-like
    abstraction.

    Args:
      path (PathLike): Pathlike object.
    
    Returns:
      A pathlib-like abstraction.
    """
    if isinstance(path, str):
        uri_splits = path.split('://', maxsplit=1)
        if len(uri_splits) > 1:    
            # str is URI (e.g. `gs://`, `github://`,...)
            return _PREFIXES_TO_CLS[f'{uri_splits[0]}://'](path)
        return FilePath(path)
    elif isinstance(path, _PATHLIKE_CLS):
        return path
    elif isinstance(path, os.PathLike):
        return FilePath(path)
    else: raise TypeError(f'Invalid path type: {path!r}')


def get_userhome(as_pathz: bool = True):
    h = os.path.expanduser('~')
    return as_path(h) if as_pathz else h

def get_cwd():
    return os.getcwd()


def resolve_relative(filepath: PathLike) -> str:
    """
    If the filepath is a relative path, convert it to an absolute path
    
    Args:
      filepath (PathLike): The path to the file you want to resolve.
    
    Returns:
      A string.
    """
    if not isinstance(filepath, str): filepath = filepath.as_posix()
    if '://' in filepath: return filepath
    if filepath.startswith('~'): filepath = filepath.replace('~', get_userhome(), 1)
    elif filepath.startswith('../'): filepath = filepath.replace('..', get_cwd(), 1)
    elif filepath.startswith('..'): filepath = filepath.replace('..', f'{pathlib.Path(get_cwd()).parent.parent.as_posix()}/', 1)
    elif filepath.startswith('./'): filepath = filepath.replace('.', get_cwd(), 1)
    elif filepath.startswith('.'): filepath = filepath.replace('.', f'{pathlib.Path(get_cwd()).parent.as_posix()}/', 1)
    return filepath

def get_path(filepath: PathLike, resolve: bool = False) -> FileLike:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath

def get_pathlike(filepath: PathLike, resolve: bool = False) -> FileLike:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath

to_path = get_path

def get_filelike(path: FileType) -> FileLike:
    #if isinstance(file, PathzLike): return file
    if hasattr(path, '_cloudstr'): return path
    if hasattr(path, 'as_posix'): return get_path(path.as_posix())
    if isinstance(path, str): return get_path(path)
    if hasattr(path, 'file') and hasattr(getattr(path, 'file'), 'name'): return get_path(path.file.name)
    return get_path(path.name) if hasattr(path, 'name') else path

# FileT = TypeVar('FileT', bound = FileLike)

class File(CloudFileSystemPath):
    
    # __metaclass__ = Type[FileLike]

    def __new__(
        cls, 
        *args, 
        load_file: Optional[bool] = False, 
        mode: LoadMode = 'default', 
        loader: Callable = None, 
        **kwargs
    ) -> FileLikeT:

        _file = get_filelike(*args, **kwargs)
        if load_file: return cls.load_file(file = _file, mode = mode, loader = loader)
        return _file

    @classmethod
    def get_tempfile(
        cls,
        *args,
        suffix: Optional[str] = None, 
        prefix: Optional[str] = None,  
        dir: Optional[str] = None,
        delete: bool = False,
        delete_on_exit: Optional[bool] = True,
        **kwargs
    ) -> FileLike:
        """
        Creates a new temporary file
        """
        f: tempfile._TemporaryFileWrapper = tempfile.NamedTemporaryFile(*args, suffix = suffix, prefix = prefix, dir = dir, delete = delete, **kwargs)
        f.close()
        p = get_path(f.name)
        if delete_on_exit:
            atexit.register(p.unlink, missing_ok = True)
        return p
    
    @classmethod
    def get_tempdir(
        cls,
        suffix: Optional[str] = None, 
        prefix: Optional[str] = None,  
        dir: Optional[str] = None, 
        delete_on_exit: Optional[bool] = False,
        **kwargs
    ) -> FileLike:
        """
        Creates a new temporary directory
        """
        d = tempfile.TemporaryDirectory(suffix = suffix, prefix = prefix, dir = dir)
        p = get_path(d.name)
        if delete_on_exit:
            atexit.register(p.rmdir, missing_ok = True)
        return p


    @classmethod
    async def async_load_json(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        **kwargs
    ) -> Union[Dict[Any, Any], List[Any], Any]:
        _file = get_filelike(*args, **kwargs) if file is None else file
        from fileio.io import Json
        return await ThreadPooler.run_async(
            Json.loads,
            await _file.async_read_text()
        )
        # return Json.loads(await _file.async_read_text())
    
    @classmethod
    async def async_load_yaml(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        **kwargs
    ) -> Union[Dict[Any, Any], List[Any], Any]:
        _file = get_filelike(*args, **kwargs) if file is None else file
        from fileio.io import Yaml
        return await ThreadPooler.run_async(
            Yaml.loads,
            await _file.async_read_text()
        )
        # return Yaml.loads(await _file.async_read_text())
    
    @classmethod
    async def async_load_text(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        **kwargs
    ) -> str:
        _file = get_filelike(*args, **kwargs) if file is None else file
        return await _file.async_read_text()
    
    @classmethod
    async def async_load_pickle(
        cls, 
        *args, file: Union[FileLike, Any] = None, 
        **kwargs
    ) -> Any:
        _file = get_filelike(*args, **kwargs) if file is None else file
        from fileio.io import Dill
        return await ThreadPooler.run_async(
            Dill.loads,
            await _file.async_read_bytes()
        )
        # return Dill.loads(await _file.async_read_bytes())
    
    @classmethod
    async def async_load_csv(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        csv_args: Dict[str, Any] = None,
        **kwargs
    ) -> Any:
        _file = get_filelike(*args, **kwargs) if file is None else file
        if csv_args is None: csv_args = {}
        from fileio.io import Csv
        return await ThreadPooler.run_async(
            Csv.load,
            _file,
            **csv_args
        )

    @classmethod
    async def async_load_tsv(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        tsv_args: Dict[str, Any] = None,
        **kwargs
    ) -> Any:
        _file = get_filelike(*args, **kwargs) if file is None else file
        if tsv_args is None: tsv_args = {}
        from fileio.io import Tsv
        return await ThreadPooler.run_async(
            Tsv.load,
            _file,
            **tsv_args
        )

    
    @classmethod
    async def async_load_file(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        mode: LoadMode = 'default', 
        loader: Callable = None, 
        **kwargs
    ) -> Any:
        _file = get_filelike(*args, **kwargs) if file is None else file
        if loader is not None:
            data = await _file.async_read_bytes() if mode == LoadMode.binary else await _file.async_read_text()
            if inspect.iscoroutinefunction(loader):
                return await loader(data)
            return loader(data)
        if _file.extension == '.json':
            return await cls.async_load_json(file = _file)
        if _file.extension in {'.yml', '.yaml'}:
            return await cls.async_load_yaml(file = _file)
        if _file.extension in {'.pickle', '.pkl'}:
            return await cls.async_load_pickle(file = _file)
        if _file.extension in {'.txt', '.text'}:
            return await cls.async_load_text(file = _file)
        if _file.extension in {'.csv'}:
            return await cls.async_load_csv(file = _file)
        if _file.extension in {'.tsv'}:
            return await cls.async_load_tsv(file = _file)
        raise ValueError(f'Unknown file extension: {_file.extension}')
    
    @classmethod
    def load_json(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        **kwargs
    ) -> Union[Dict[Any, Any], List[Any], Any]:
        _file = get_filelike(*args, **kwargs) if file is None else file
        from fileio.io import Json
        return Json.loads(_file.read_text())
    
    @classmethod
    def load_yaml(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        **kwargs
    ) -> Union[Dict[Any, Any], List[Any], Any]:
        _file = get_filelike(*args, **kwargs) if file is None else file
        from fileio.io import Yaml
        return Yaml.loads(_file.read_text())
    
    @classmethod
    def load_text(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        **kwargs
    ) -> str:
        _file = get_filelike(*args, **kwargs) if file is None else file
        return _file.read_text()
    
    @classmethod
    def load_pickle(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        **kwargs
    ) -> Any:
        _file = get_filelike(*args, **kwargs) if file is None else file
        from fileio.io import Dill
        return Dill.loads(_file.read_bytes())
    
    @classmethod
    def load_csv(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        csv_args: Dict[str, Any] = None,
        **kwargs
    ) -> Any:
        _file = get_filelike(*args, **kwargs) if file is None else file
        if csv_args is None: csv_args = {}
        from fileio.io import Csv
        return Csv.load(_file, **csv_args)

    @classmethod
    def load_tsv(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        tsv_args: Dict[str, Any] = None,
        **kwargs
    ) -> Any:
        _file = get_filelike(*args, **kwargs) if file is None else file
        if tsv_args is None: tsv_args = {}
        from fileio.io import Tsv
        return Tsv.load(_file, **tsv_args)

    @classmethod
    def load_file(
        cls, 
        *args, 
        file: Union[FileLike, Any] = None, 
        mode: LoadMode = 'default', 
        loader: Callable = None, 
        **kwargs
    ) -> Any:
        _file = get_filelike(*args, **kwargs) if file is None else file
        if loader is not None:
            data = _file.read_bytes() if mode == LoadMode.binary else _file.read_text()
            return loader(data)
        if _file.extension == '.json':
            return cls.load_json(file = _file)
        if _file.extension in {'.yml', '.yaml'}:
            return cls.load_yaml(file = _file)
        if _file.extension in {'.pickle', '.pkl'}:
            return cls.load_pickle(file = _file)
        if _file.extension in {'.txt', '.text'}:
            return cls.load_text(file = _file)
        if _file.extension in {'.csv'}:
            return cls.load_csv(file = _file)
        if _file.extension in {'.tsv'}:
            return cls.load_tsv(file = _file)
        raise ValueError(f'Unknown file extension: {_file.extension}')
    
    # Pydantic methods
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[FileLike, Any]) -> FileLike:
        return get_filelike(v) if v is not None else None
    
    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        field_schema.update(
            type='string',
            format='binary',
        )
    



class StatelessFile:
    
    """
    Stateless file are written to a local file during transactions
    and then uploaded to the remote storage once the transaction is complete.
    """

    def __init__(
        self, 
        input_file:  Optional[FileLike] = None,
        output_file: Optional[FileLike] = None,
        overwrite: Optional[bool] = False,
        load_file: Optional[bool] = False, 
        mode: LoadMode = 'default', 
        loader: Callable = None, 
        enable_auto_filename: Optional[bool] = None,
        output_file_suffix: Optional[str] = None,
        **kwargs
    ):
        if input_file:
            input_file = get_filelike(input_file, **kwargs)
            if load_file: input_file = File.load_file(file = input_file, mode = mode, loader = loader)
        self.file: FileLike = input_file if input_file is not None else None
        self._enable_auto_filename = enable_auto_filename
        self._output_file_suffix = output_file_suffix
        
        self._temp_src_file = File.get_tempfile(delete_on_exit = False)
        self._temp_out_file = File.get_tempfile(delete_on_exit = False)

        # self.output_file = File(output_file) if output_file is not None else None
        self._prepare_output_file(output_file = output_file)
        self.overwrite = overwrite
        # self._write_ready: bool = False
        self._read_ready: bool = False
        self._ready: bool = self.output_file is not None
        self._closed: bool = False
        atexit.register(self._onexit)
    
    def _create_autofile(self, file: FileLike, suffix: Optional[str] = None, parent: Optional[FileLike] = None):
        """
        Creates an auto filename for the given file.
        """
        suffix = suffix or self._output_file_suffix or file.suffix
        file_dir: FileLike = parent or (file.parent if file.is_file() else file)
        for i in range(100):
            output_file = file_dir.joinpath(f'{file.stem}_v{str(i).zfill(3)}{suffix}')
            if not output_file.exists():
                break
        return output_file

    def _prepare_output_file(self, output_file: Optional[FileLike] = None):
        """
        Prepares the output file for writing.
        - designed to handle edge cases where the output_file param
          is relative, is a directory, or is a file that already exists.
        """
        self.output_file: FileLike = None
        if not output_file: 
            if self.file and self._enable_auto_filename:
                self.output_file = self._create_autofile(file = self.file)
            return
        output_file = File(output_file)
        if output_file.is_dir() and self._enable_auto_filename:
            self.output_file = self._create_autofile(file = (self.file or output_file), parent = output_file)

        # if output_file.exists() and self._enable_auto_filename:
        #     # we'll create iterated filenames until we find one that doesn't exist
        #     self.output_file = self._create_autofile(file = output_file)

        elif not output_file.is_file():
            if self.file and self._enable_auto_filename:
                self.output_file = self._create_autofile(file = self.file, parent = output_file)
            elif self.file:
                output_file = (
                    output_file.joinpath(
                        f'{self.file.stem}_output{self._output_file_suffix or self.file.suffix}'
                    )
                    if output_file.joinpath(self.file.stem + (self._output_file_suffix or self.file.suffix)).exists()
                    else output_file.joinpath(self.file.stem + (self._output_file_suffix or self.file.suffix))
                )
            else:
                self.output_file = output_file.joinpath(f'{self._temp_out_file.name}{".output" if self._output_file_suffix is None else self._output_file_suffix}')
        else:
            self.output_file = output_file
        # self._write_ready = True
    
    def _prepare_file(self):
        """
        Function to prepare the file for read/writing.
        """
        if not self._read_ready:
            if self.file:
                self._temp_src_file.write_bytes(self.file.read_bytes())
            self._read_ready = True
        
    async def _async_prepare_file(self):
        """
        Function to prepare the file for read/writing.
        """
        if not self._read_ready:
            if self.file:
                await self._temp_src_file.async_write_bytes(await self.file.async_read_bytes())
            self._read_ready = True

    def write_bytes(self, data, flush: Optional[bool] = False, **kwargs):
        self._temp_out_file.write_bytes(data)
        if flush: self.flush(**kwargs)

    async def async_write_bytes(self, data: bytes, flush: Optional[bool] = False, **kwargs):
        await self._temp_out_file.async_write_bytes(data)
        if flush: await self.async_flush(**kwargs)

    def write_text(self, data, flush: Optional[bool] = False, **kwargs):
        self._temp_out_file.write_text(data)
        if flush: self.flush(**kwargs)

    async def async_write_text(self, data: str, flush: Optional[bool] = False, **kwargs):
        await self._temp_out_file.async_write_text(data, **kwargs)
        if flush: await self.async_flush(**kwargs)

    def write(self, data: Union[str, bytes], mode: str = 'w', flush: Optional[bool] = False, **kwargs):
        with self._temp_out_file.open(mode = mode, **kwargs) as f:
            f.write(data)
        if flush: self.flush(**kwargs)
    
    async def async_write(self, data: Union[str, bytes], mode: str = 'w', flush: Optional[bool] = False, **kwargs):
        async with self._temp_out_file.async_open(mode = mode, **kwargs) as f:
            await f.write(data)
        if flush: await self.async_flush(**kwargs)
    
    def read_bytes(self):
        self._prepare_file()
        if self.file:
            return self._temp_src_file.read_bytes()
        return self._temp_out_file.read_bytes()
    
    async def async_read_bytes(self):
        await self._async_prepare_file()
        if self.file:
            return await self._temp_src_file.async_read_bytes()
        return await self._temp_out_file.async_read_bytes()
    
    def read_text(self):
        self._prepare_file()
        if self.file:
            return self._temp_src_file.read_text()
        return self._temp_out_file.read_text()
    
    async def async_read_text(self):
        await self._async_prepare_file()
        if self.file:
            return await self._temp_src_file.async_read_text()
        return await self._temp_out_file.async_read_text()

    def read(self, mode: str = 'r', **kwargs):
        self._prepare_file()
        if self.file:
            with self._temp_src_file.open(mode = mode, **kwargs) as f:
                return f.read()
        with self._temp_out_file.open(mode = mode, **kwargs) as f:
            return f.read()
        
    async def async_read(self, mode: str = 'r', **kwargs):
        await self._async_prepare_file()
        async with self._temp_src_file.async_open(mode = mode, **kwargs) as f:
            return await f.read()

    def open(self, mode: str = 'r', **kwargs):
        if 'r' in mode and self.file:
            self._prepare_file()
            return self._temp_src_file.open(mode = mode, **kwargs)
        return self._temp_out_file.open(mode = mode, **kwargs)

    def async_open(self, mode: str = 'r', **kwargs) -> IterableAIOFile:
        if 'r' in mode and self.file:
            self._prepare_file()
            # logger.info('Opening: tmp src')
            return self._temp_src_file.async_open(mode = mode, **kwargs)
        # logger.info('Opening: tmp out')
        return self._temp_out_file.async_open(mode = mode, **kwargs)

    @lazyproperty
    def text(self):
        """
        Returns the text of the temp src file
        """
        self._prepare_file()
        if self.file:
            return self._temp_src_file.read_text()
        return self._temp_out_file.read_text()
    
    @lazyproperty
    def output_text(self):
        """
        Returns the text of the temp out file
        """
        if self._closed:
            if self.output_file:
                return self.output_file.read_text()
            raise RuntimeError('Cannot read output text after closing the file, and no output file was provided.')
        return self._temp_out_file.read_text()

    @lazyproperty
    def bytes(self):
        """
        Returns the bytes of the temp src file
        """
        self._prepare_file()
        if self.file:
            return self._temp_src_file.read_bytes()
        return self._temp_out_file.read_bytes()
    
    @lazyproperty
    def output_bytes(self):
        """
        Returns the bytes of the temp out file
        """
        if self._closed:
            if self.output_file:
                return self.output_file.read_bytes()
            raise RuntimeError('Cannot read output bytes after closing the file, and no output file was provided.')
        return self._temp_out_file.read_bytes()

    @lazyproperty
    async def async_text(self):
        """
        Returns the text of the temp src file
        """
        await self._async_prepare_file()
        if self.file:
            return await self._temp_src_file.async_read_text()
        return await self._temp_out_file.async_read_text()
    
    @lazyproperty
    async def async_output_text(self):
        """
        Returns the text of the temp out file
        """
        if self._closed:
            if self.output_file:
                return await self.output_file.async_read_text()
            raise RuntimeError('Cannot read output text after closing the file, and no output file was provided.')
        return await self._temp_out_file.async_read_text()

    @lazyproperty
    async def async_bytes(self):
        """
        Returns the bytes of the temp src file
        """
        await self._async_prepare_file()
        if self.file:
            return await self._temp_src_file.async_read_bytes()
        return await self._temp_out_file.async_read_bytes()

    @lazyproperty
    async def async_output_bytes(self):
        """
        Returns the bytes of the temp out file
        """
        if self._closed:
            if self.output_file:
                return await self.output_file.async_read_bytes()
            raise RuntimeError('Cannot read output bytes after closing the file, and no output file was provided.')
        return await self._temp_out_file.async_read_bytes()

    @lazyproperty
    def path(self):
        """
        Returns the path of the temp src file
        """
        self._prepare_file()
        if self.file:
            return self._temp_src_file.as_posix()
        return self._temp_out_file.as_posix()
    
    @lazyproperty
    def input_path(self):
        """
        Returns the path of the temp src file
        """
        if self._closed:
            raise RuntimeError('File is closed')
        self._prepare_file()
        return self._temp_src_file.as_posix()
    
    @lazyproperty
    def read_path(self) -> str:
        """
        Returns the path of the temp src file
        """
        if self._closed:
            raise RuntimeError('File is closed')
        self._prepare_file()
        return self._temp_src_file.as_posix()
    
    @lazyproperty
    def write_path(self) -> str:
        """
        Returns the path of the temp output file
        """
        if self._closed:
            raise RuntimeError('File is closed')
        return self._temp_out_file.as_posix()

    @lazyproperty
    def source_path(self):
        """
        Returns the path of the temp src file
        """
        return self.file.as_posix() if self.file else self._temp_src_file.as_posix()

    @lazyproperty
    def target_path(self):
        """
        Returns the path of the target output file
        """
        if self.output_file is None:
            if self.file and self.file.exists():
                return (
                    self.file.as_posix()
                    if self.overwrite
                    else self._temp_out_file.as_posix()
                )
            return self.file.as_posix()
        return self.output_file.as_posix()
    

    @lazyproperty
    def target_file(self) -> 'FileLike':
        """
        Returns the path of the target output file
        """
        if self.output_file is None:
            if self.file and self.file.exists():
                return (
                    self.file
                    if self.overwrite
                    else self._temp_out_file
                )
            return self.file
        return self.output_file

    def __getattr__(self, name, default: Any = None):
        self._prepare_file()
        return getattr(self._temp_src_file, name, default)
    
    def __getitem__(self, key):
        self._prepare_file()
        return self._temp_src_file[key]
    
    def __call__(self, *args, **kwargs):
        self._prepare_file()
        return self._temp_src_file(*args, **kwargs)

    def flush(
        self, 
        overwrite: Optional[bool] = None, 
        output_file: Optional[FileLike] = None, 
        **kwargs
    ):
        """
        Flushes the file to the output file.
        """
        if self._closed: 
            logger.warning('File is closed. Cannot flush.')
            return
        overwrite = overwrite or self.overwrite
        output_file = output_file or self.output_file
        if output_file is not None:
            logger.info(f'Writing to output file: {output_file}')
            output_file.write_bytes(self._temp_out_file.read_bytes())
        elif self.file and (overwrite or not self.file.exists()):
            logger.info(f'Writing to input file: {self.file}')
            self.file.write_bytes(self._temp_out_file.read_bytes())
    
    async def async_flush(
        self, 
        overwrite: Optional[bool] = None, 
        output_file: Optional[FileLike] = None, 
        **kwargs
    ):
        """
        Flushes the file to the output file.
        """
        if self._closed: 
            logger.warning('File is closed. Cannot flush.')
            return
        overwrite = overwrite or self.overwrite
        output_file = output_file or self.output_file
        if output_file is not None:
            logger.info(f'Writing to output file: {output_file}')
            await output_file.async_write_bytes(await self._temp_out_file.async_read_bytes())
        elif self.file and (overwrite or not self.file.exists()):
            logger.info(f'Writing to input file: {self.file}')
            await self.file.async_write_bytes(await self._temp_out_file.async_read_bytes())


    def close(self, *args, overwrite: Optional[bool] = None, output_file: Optional[FileLike] = None, **kwargs):
        if self._closed:
            return
        if self._ready:
            self.flush(overwrite = overwrite, output_file = output_file, **kwargs)
            self._ready = False        
        self._temp_src_file.rm_file(missing_ok=True)
        self._temp_out_file.rm_file(missing_ok=True)
        self._closed = True
    
    async def async_close(self, *args, overwrite: Optional[bool] = None, output_file: Optional[FileLike] = None, **kwargs):
        if self._closed:
            return
        if self._ready:
            await self.async_flush(overwrite = overwrite, output_file = output_file, **kwargs)
            self._ready = False
        
        await self._temp_src_file.async_rm_file(missing_ok=True)
        await self._temp_out_file.async_rm_file(missing_ok=True)
        self._closed = True

    def __enter__(self):
        self._prepare_file()
        return self._temp_src_file
    
    def __exit__(self, *args):
        self.close()
    
    def _onexit(self, *args, **kwargs):
        try:
            self.close()
        except Exception as e:
            pass
            # logger.error(e)
    
    def dict(self, *args, **kwargs):
        # self.flush()
        return {
            'input_path': self.input_path,
            'output_path': self.target_path,
            'source_path': self.source_path,
        }


__all__ = (
    'get_path', 
    'as_path', 
    'PathLike', 
    'FilePath', 
    'FileS3Path', 
    'FileGSPath', 
    'FileLike',
    'FileS3CPath',
)


from fileio.utils import logger

# MPU seems to work > 150MB?
_debug_mode: bool = False

def _log(f, *args, verbose: bool = False, **kwargs):
    if _debug_mode or verbose:
        logger.info(f, *args, **kwargs)
    else:
        logger.debug(f, *args, **kwargs)




from __future__ import annotations

"""
Azure Provider for FileIO
"""

import os

from fileio.lib.posix.base import *
from fileio.lib.posix.filesys import Azure_CloudFileSystem
from fileio.lib.posix.cloud import *

class FileAZPurePath(CloudFileSystemPurePath):
    _prefix: str = 'az'
    _provider: str = 'Azure'
    _win_pathz: ClassVar = 'PureFileAZWindowsPath'
    _posix_pathz: ClassVar = 'PureFileAZPosixPath'


class PureFileAZPosixPath(PureCloudFileSystemPosixPath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_posix_flavour
    _pathlike = posixpath
    __slots__ = ()


class PureFileAZWindowsPath(PureCloudFileSystemWindowsPath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_windows_flavour
    _pathlike = ntpath
    __slots__ = ()

class FileAZPath(CloudFileSystemPath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _pathz_windows_flavour if os.name == 'nt' else _pathz_posix_flavour
    _accessor: AccessorLike = None
    _pathlike = posixpath
    _prefix = 'az'
    _provider = 'Azure'

    _win_pathz: ModuleType = 'FileAZWindowsPath'
    _posix_pathz: ModuleType = 'FileAZPosixPath'

    def _init(self, template: Optional['FileAZPath'] = None):
        self._accessor: AccessorLike = FileSysManager.get_accessor(self._prefix)
        # self._accessor: AccessorLike = get_accessor(self._prefix)
        self._closed = False
        self._fileio = None

    def __new__(cls, *parts, **kwargs):
        if cls is FileAZPath or issubclass(cls, FileAZPath): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self
    

class FileAZPosixPath(PosixPath, FileAZPath, PureFileAZPosixPath):
    __slots__ = ()


class FileAZWindowsPath(WindowsPath, FileAZPath, PureFileAZWindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("FileAZPath.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("FileAZPath.async_is_mount() is unsupported on this system")

register_pathlike(
    [
        FileAZPurePath, FileAZPath, PureFileAZPosixPath, FileAZWindowsPath, FileAZPosixPath, PureFileAZWindowsPath
    ]
)

AZFileSystem = Azure_CloudFileSystem


__all__ = (
    'FileAZPurePath',
    'FileAZPath',
    'PureFileAZPosixPath',
    'FileAZWindowsPath',
    'FileAZPosixPath',
    'PureFileAZWindowsPath',
    'AZFileSystem'
)



from __future__ import absolute_import

from fileio.utils.logs import default_logger as logger
from fileio.utils.lazylib import LazyLib

from fileio.utils.helpers import *
from fileio.utils.configs import settings

from fileio.utils.ops import (
    get_url_file_name, 
    checksum_file,
    async_checksum_file,
    fetch_file_from_url,
    async_fetch_file_from_url,
)


import os
import json
import pathlib
import multiprocessing as mp
from pydantic import BaseSettings, validator, root_validator

from typing import Optional, Dict, Any, Union
from fileio.utils.logs import default_logger as logger
from fileio.types.classprops import lazyproperty

class CoreSettings(BaseSettings):

    authz_config_dir: Optional[pathlib.Path] = None
    boto_config: Optional[pathlib.Path] = None
    max_workers: Optional[int] = None

    class Config:
        env_prefix: str = ""

    @validator("authz_config_dir", pre=True)
    def validate_authz_config_dir(cls, v):
        return pathlib.Path("~/.auth").expanduser() if v is None else v
    
    @lazyproperty
    def num_workers(self):
        return self.max_workers or min(2, round(mp.cpu_count() // 2))

    @lazyproperty
    def boto_config_path(self) -> pathlib.Path:
        if self.boto_config is None:
            return pathlib.Path('/root/.boto') if self.in_colab else pathlib.Path("~/.boto").expanduser()
        return self.boto_config

    @lazyproperty
    def boto_config_exists(self):
        return self.boto_config_path.exists()

    @lazyproperty
    def user_home(self) -> pathlib.Path:
        if self.in_colab:
            return pathlib.Path("/content")
        return pathlib.Path("~").expanduser()
    
    @lazyproperty
    def in_colab(self) -> bool:
        try:
            from google.colab import drive
            return True
        except ImportError:
            return False
    
    def set_env(self):
        if self.boto_config_exists:
            os.environ["BOTO_CONFIG"] = self.boto_config_path.as_posix()
            os.environ["BOTO_PATH"] = self.boto_config_path.as_posix()
    
    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k):  continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)


core_settings = CoreSettings()

class AwsSettings(BaseSettings):
    aws_access_token: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_region: Optional[str] = "us-east-1"
    set_s3_endpoint: Optional[bool] = False
    s3_config: Optional[Union[str, Dict[str, Any]]] = None

    class Config:
        env_prefix: str = ""

    @validator("s3_config", pre=True)
    def validate_s3_config(cls, v):
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v
    
    @lazyproperty
    def s3_endpoint(self):
        return f'https://s3.{self.aws_region}.amazonaws.com'
    
    def set_env(self):
        if self.aws_access_key_id:
            os.environ['AWS_ACCESS_KEY_ID'] = self.aws_access_key_id
        if self.aws_secret_access_key:
            os.environ['AWS_SECRET_ACCESS_KEY'] = self.aws_secret_access_key
        if self.aws_region:
            os.environ['AWS_REGION'] = self.aws_region
        if self.aws_access_token:
            os.environ['AWS_ACCESS_TOKEN'] = self.aws_access_token
        if self.set_s3_endpoint:
            os.environ['S3_ENDPOINT'] = self.s3_endpoint
    

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k):  continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_auth(self, update_fs: bool = True, **config):
        self.update_config(**config)
        self.set_env()

        if update_fs:
            from fileio.lib.posix.filesys import FileSysManager
            FileSysManager.get_accessor('s3', _reset=True)

    def build_s3fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config = {}
        if self.aws_access_key_id:
            config["key"] = self.aws_access_key_id
            # config['aws_access_key_id'] = self.aws_access_key_id
        if self.aws_secret_access_key:
            config["secret"] = self.aws_secret_access_key
            # config['aws_secret_access_key'] = self.aws_secret_access_key
        if self.aws_access_token:
            config["token"] = self.aws_access_token
        if not (config.get('key') and config.get('secret')) and not core_settings.boto_config_exists:
            config['anon'] = True
        if self.set_s3_endpoint:
            config['client_kwargs'] = {'endpoint_url': self.s3_endpoint, 'region_name': self.aws_region}
        if self.s3_config:
            config['config_kwargs'] = self.s3_config
        
        return config


class GcpSettings(BaseSettings):
    gcp_project: Optional[str] = None
    gcloud_project: Optional[str] = None
    google_cloud_project: Optional[str] = None
    google_application_credentials: Optional[pathlib.Path] = None

    gcs_client_config: Optional[Union[str, Dict[str, Any]]] = None
    gcs_config: Optional[Union[str, Dict[str, Any]]] = None

    class Config:
        env_prefix: str = ""

    @validator("google_application_credentials")
    def validate_google_application_credentials(cls, v):
        return core_settings.user_home.joinpath('adc.json') if v is None else v

    @validator("gcs_client_config")
    def validate_gcs_client_config(cls, v) -> Dict:
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v
    
    @validator("gcs_config")
    def validate_gcs_config(cls, v) -> Dict:
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v

    @lazyproperty
    def adc_exists(self):
        return self.google_application_credentials.exists()
    
    @lazyproperty
    def project(self):
        return self.gcp_project or self.gcloud_project or self.google_cloud_project
    
    def set_env(self):
        if self.adc_exists:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.google_application_credentials.as_posix()
        if self.project:
            os.environ["GOOGLE_CLOUD_PROJECT"] = self.project
    
    def build_gcsfs_config(self) -> Dict[str, Any]:
        """
        Builds the gcsfs config kwargs
        """
        config = {}
        if self.adc_exists: config['token'] = self.google_application_credentials.as_posix()
        if self.project: config['project'] = self.project
        if self.gcs_client_config: config['client_config'] = self.gcs_client_config
        if self.gcs_config: config['config_kwargs'] = self.gcs_config
        return config

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k):  continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_auth(self, update_fs: bool = True, **config):
        self.update_config(**config)
        self.set_env()

        if update_fs:
            # Reset the accessor to use the new settings
            from fileio.lib.posix.filesys import FileSysManager
            FileSysManager.get_accessor('gs', _reset=True)

    

class MinioSettings(BaseSettings):
    minio_endpoint: Optional[str] = None
    minio_access_key: Optional[str] = None
    minio_secret_key: Optional[str] = None
    minio_access_token: Optional[str] = None
    minio_secure: Optional[bool] = True
    minio_region: Optional[str] = None
    minio_config: Optional[Union[str, Dict[str, Any]]] = None
    minio_signature_ver: Optional[str] = 's3v4'

    class Config:
        env_prefix: str = ""

    @validator("minio_config", pre=True)
    def validate_minio_config(cls, v):
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v
    
    def set_env(self):
        if self.minio_endpoint:
            os.environ["MINIO_ENDPOINT"] = self.minio_endpoint
        if self.minio_access_key:
            os.environ["MINIO_ACCESS_KEY"] = self.minio_access_key
        if self.minio_secret_key:
            os.environ["MINIO_SECRET_KEY"] = self.minio_secret_key
        if self.minio_secure:
            os.environ["MINIO_SECURE"] = str(self.minio_secure)
        if self.minio_region:
            os.environ["MINIO_REGION"] = self.minio_region
        if self.minio_signature_ver:
            os.environ["MINIO_SIGNATURE_VER"] = self.minio_signature_ver


    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k):  continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_auth(self, update_fs: bool = True, **config):
        self.update_config(**config)
        self.set_env()

        # Reset the accessor to use the new settings
        from fileio.lib.posix.filesys import FileSysManager
        FileSysManager.get_accessor('minio', _reset=True)
    
    def build_s3fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config = {
            "client_kwargs": {
                "endpoint_url": self.minio_endpoint,
                "region_name": self.minio_region,
            },
            "config_kwargs": {
                "signature_version": self.minio_signature_ver,
            }
        }
        if self.minio_access_key:
            config["key"] = self.minio_access_key
        if self.minio_secret_key:
            config["secret"] = self.minio_secret_key
        if self.minio_access_token:
            config["token"] = self.minio_access_token
        if self.minio_config:
            config["config_kwargs"].update(self.minio_config)
        return config

class S3CompatSettings(BaseSettings):
    s3_compat_endpoint: Optional[str] = None
    s3_compat_access_key: Optional[str] = None
    s3_compat_secret_key: Optional[str] = None
    s3_compat_access_token: Optional[str] = None
    s3_compat_secure: Optional[bool] = True
    s3_compat_region: Optional[str] = None
    s3_compat_config: Optional[Union[str, Dict[str, Any]]] = None
    s3_compat_signature_ver: Optional[str] = 's3v4'

    class Config:
        env_prefix: str = ""

    @validator("s3_compat_config")
    def validate_s3_compat_config(cls, v) -> Dict:
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v
    
    def set_env(self):
        if self.s3_compat_endpoint:
            os.environ["S3_COMPAT_ENDPOINT"] = self.s3_compat_endpoint
        if self.s3_compat_access_key:
            os.environ["S3_COMPAT_ACCESS_KEY"] = self.s3_compat_access_key
        if self.s3_compat_secret_key:
            os.environ["S3_COMPAT_SECRET_KEY"] = self.s3_compat_secret_key
        if self.s3_compat_access_token:
            os.environ["S3_COMPAT_ACCESS_TOKEN"] = self.s3_compat_access_token
        if self.s3_compat_secure:
            os.environ["S3_COMPAT_SECURE"] = str(self.s3_compat_secure)
        if self.s3_compat_region:
            os.environ["S3_COMPAT_REGION"] = self.s3_compat_region

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_auth(self, **config):
        self.update_config(**config)
        self.set_env()
    

    def build_s3fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config = {
            "client_kwargs": {
                "endpoint_url": self.s3_compat_endpoint,
                "region_name": self.s3_compat_region,
            },
            "config_kwargs": {
                "signature_version": self.s3_compat_signature_ver,
            }
        }
        if self.s3_compat_access_key:
            config["key"] = self.s3_compat_access_key
        if self.s3_compat_secret_key:
            config["secret"] = self.s3_compat_secret_key
        if self.s3_compat_access_token:
            config["token"] = self.s3_compat_access_token
        if self.s3_compat_config:
            config["config_kwargs"].update(self.s3_compat_config)
        return config


class AzureSettings(BaseSettings):
    azure_account_name: Optional[str] = None
    azure_account_key: Optional[str] = None
    azure_sas_token: Optional[str] = None
    azure_connection_string: Optional[str] = None
    azure_blocksize: Optional[int] = None
    azure_client_id: Optional[str] = None
    azure_client_secret: Optional[str] = None
    azure_tenant_id: Optional[str] = None
    azure_version_aware: Optional[bool] = False

    azure_secure: Optional[bool] = True
    azure_config: Optional[Union[str, Dict[str, Any]]] = None
    azure_socket_timeout: Optional[int] = None

    class Config:
        env_prefix: str = ""

    @validator("azure_config")
    def validate_azure_config(cls, v) -> Dict:
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v
    
    def set_env(self):
        if self.azure_account_name:
            os.environ["AZURE_STORAGE_ACCOUNT_NAME"] = self.azure_account_name
        if self.azure_account_key:
            os.environ["AZURE_STORAGE_ACCOUNT_KEY"] = self.azure_account_key
        if self.azure_sas_token:
            os.environ["AZURE_STORAGE_SAS_TOKEN"] = self.azure_sas_token
        if self.azure_connection_string:
            os.environ["AZURE_STORAGE_CONNECTION_STRING"] = self.azure_connection_string
        if self.azure_client_id:
            os.environ["AZURE_STORAGE_CLIENT_ID"] = self.azure_client_id
        if self.azure_client_secret:
            os.environ["AZURE_STORAGE_CLIENT_SECRET"] = self.azure_client_secret
        if self.azure_tenant_id:
            os.environ["AZURE_STORAGE_TENANT_ID"] = self.azure_tenant_id
        

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_auth(self, **config):
        self.update_config(**config)
        self.set_env()
    
    def build_azurefs_config(self) -> Dict[str, Any]:
        """
        Builds the azurefs config dict
        """
        config = {
            "account_name": self.azure_account_name,
            "account_key": self.azure_account_key,
            "sas_token": self.azure_sas_token,
            "secure": self.azure_secure,
            "connection_string": self.azure_connection_string,
            "blocksize": self.azure_blocksize,
            "client_id": self.azure_client_id,
            "client_secret": self.azure_client_secret,
            "tenant_id": self.azure_tenant_id,
            "version_aware": self.azure_version_aware,
            "socket_timeout": self.azure_socket_timeout,
        }
        config = {k: v for k, v in config.items() if v is not None}
        if self.azure_config:
            config.update(self.azure_config)
        return config

class GithubSettings(BaseSettings):
    github_org: Optional[str] = None
    github_repo: Optional[str] = None
    github_user: Optional[str] = None
    github_token: Optional[str] = None
    github_sha: Optional[str] = None

    class Config:
        env_prefix: str = ""
        case_sensitive = False

    def set_env(self):
        if self.github_org:
            os.environ["GITHUB_ORG"] = self.github_org
        if self.github_repo:
            os.environ["GITHUB_REPO"] = self.github_repo
        if self.github_user:
            os.environ["GITHUB_USER"] = self.github_user
        if self.github_token:
            os.environ["GITHUB_TOKEN"] = self.github_token
        if self.github_sha:
            os.environ["GITHUB_SHA"] = self.github_sha

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_auth(self, **config):
        self.update_config(**config)
        self.set_env()

    def build_githubfs_config(self) -> Dict[str, Any]:
        """
        Builds the githubfs config dict
        """
        config = {
            'org': self.github_org,
            'repo': self.github_repo,
            'user': self.github_user,
            'token': self.github_token,
            'sha': self.github_sha,
        }
        return {k: v for k, v in config.items() if v is not None}

class HuggingfaceSettings(BaseSettings):

    hf_token: Optional[str] = None
    huggingface_token: Optional[str] = None
    hugging_face_hub_token: Optional[str] = None

    hf_org: Optional[str] = None
    hf_user: Optional[str] = None
    hf_repo: Optional[str] = None

    huggingface_org: Optional[str] = None
    huggingface_user: Optional[str] = None
    huggingface_repo: Optional[str] = None

    hf_repo_id: Optional[str] = None
    huggingface_repo_id: Optional[str] = None

    hf_repo_type: Optional[str] = None
    huggingface_repo_type: Optional[str] = None

    class Config:
        env_prefix: str = ""
        case_sensitive = False

    @lazyproperty
    def token(self) -> str:
        return self.hf_token or self.huggingface_token or self.hugging_face_hub_token
    
    @lazyproperty
    def org(self) -> str:
        return self.hf_org or self.huggingface_org
    
    @lazyproperty
    def user(self) -> str:
        return self.hf_user or self.huggingface_user
    
    @lazyproperty
    def repo(self) -> str:
        return self.hf_repo or self.huggingface_repo
    
    @lazyproperty
    def repo_id(self) -> str:
        return (
            self.hf_repo_id
            or self.huggingface_repo_id
            or f"{self.org or self.user}/{self.repo}"
        )

    @lazyproperty
    def repo_type(self) -> str:
        return self.hf_repo_type or self.huggingface_repo_type
    
    def set_env(self):
        if self.token:
            os.environ["HUGGING_FACE_HUB_TOKEN"] = self.token
        if self.repo_id:
            os.environ["HF_REPO_ID"] = self.repo_id

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)


class CloudflareR2Settings(BaseSettings):
    r2_account_id: Optional[str] = None
    r2_access_key_id: Optional[str] = None
    r2_secret_access_key: Optional[str] = None
    r2_access_token: Optional[str] = None
    r2_region: Optional[str] = None

    r2_endpoint: Optional[str] = None
    r2_config: Optional[Union[str, Dict[str, Any]]] = None

    class Config:
        env_prefix: str = ""

    @root_validator(pre=True)
    def build_valid_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate the values and build the valid values dict"""
        if values.get("r2_config") is not None:
            values["r2_config"] = json.loads(values["r2_config"]) if isinstance(values["r2_config"], str) else values["r2_config"]
        if values.get("r2_endpoint") is not None:
            values["r2_endpoint"] = values["r2_endpoint"].rstrip("/")
            if not values["r2_endpoint"].startswith("http"):
                values["r2_endpoint"] =  "https://" + values["r2_endpoint"]
        elif values.get("r2_account_id"):
            values["r2_endpoint"] = f"https://{values['r2_account_id']}.r2.cloudflarestorage.com"
        return values


    def set_env(self):
        if self.r2_endpoint:
            os.environ["R2_ENDPOINT"] = self.r2_endpoint
        if self.r2_account_id:
            os.environ["R2_ACCOUNT_ID"] = self.r2_account_id
        if self.r2_access_key_id:
            os.environ["R2_ACCESS_KEY_ID"] = self.r2_access_key_id
        if self.r2_secret_access_key:
            os.environ["R2_SECRET_ACCESS_KEY"] = self.r2_secret_access_key
        if self.r2_access_token:
            os.environ["R2_ACCESS_TOKEN"] = str(self.r2_access_token)

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_auth(self, **config):
        self.update_config(**config)
        self.set_env()
    

    def build_s3fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config = {
            "client_kwargs": {
                "endpoint_url": self.r2_endpoint,
                "region_name": self.r2_region,
            },
            "config_kwargs": {}
        }
        if self.r2_access_key_id:
            config["key"] = self.r2_access_key_id
        if self.r2_secret_access_key:
            config["secret"] = self.r2_secret_access_key
        if self.r2_access_token:
            config["token"] = self.r2_access_token
        if self.r2_config:
            config["config_kwargs"].update(self.r2_config)
        return config
    

class WasabiS3Settings(BaseSettings):

    wasabi_access_key_id: Optional[str] = None
    wasabi_secret_access_key: Optional[str] = None
    wasabi_access_token: Optional[str] = None

    wasabi_region: Optional[str] = 'us-east-1'

    wasabi_endpoint: Optional[str] = None
    wasabi_config: Optional[Union[str, Dict[str, Any]]] = None

    class Config:
        env_prefix: str = ""

    @root_validator(pre=True)
    def build_valid_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate the values and build the valid values dict"""
        if values.get("wasabi_config") is not None:
            values["wasabi_config"] = json.loads(values["wasabi_config"]) if isinstance(values["wasabi_config"], str) else values["wasabi_config"]
        if values.get("wasabi_endpoint") is not None:
            values["wasabi_endpoint"] = values["wasabi_endpoint"].rstrip("/")
            if not values["wasabi_endpoint"].startswith("http"):
                values["wasabi_endpoint"] =  "https://" + values["wasabi_endpoint"]
        else:
            values["wasabi_endpoint"] = f"https://s3.{values['wasabi_region']}.wasabisys.com" if values.get("wasabi_region", "") != "us-east-1" else \
                "https://s3.wasabisys.com"
        return values


    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_auth(self, **config):
        self.update_config(**config)
        self.set_env()
    

    def set_env(self):
        if self.wasabi_endpoint:
            os.environ["WASABI_ENDPOINT"] = self.wasabi_endpoint
        if self.wasabi_access_key_id:
            os.environ["WASABI_ACCESS_KEY_ID"] = self.wasabi_access_key_id
        if self.wasabi_secret_access_key:
            os.environ["WASABI_SECRET_ACCESS_KEY"] = self.wasabi_secret_access_key
        if self.wasabi_access_token:
            os.environ["WASABI_ACCESS_TOKEN"] = self.wasabi_access_token
        if self.wasabi_region:
            os.environ["WASABI_REGION"] = self.wasabi_region

    def build_s3fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config = {
            "client_kwargs": {
                "endpoint_url": self.wasabi_endpoint,
                "region_name": self.wasabi_region,
            },
            "config_kwargs": {}
        }
        if self.wasabi_access_key_id:
            config["key"] = self.wasabi_access_key_id
        if self.wasabi_secret_access_key:
            config["secret"] = self.wasabi_secret_access_key
        if self.wasabi_access_token:
            config["token"] = self.wasabi_access_token
        if self.wasabi_config:
            config["config_kwargs"].update(self.wasabi_config)
        return config


class Settings(BaseSettings):

    read_chunk_size: Optional[int] = 1024 * 64 # 64KB
    url_chunk_size: Optional[int] = 1024 * 128 # 128KB
    
    num_workers: Optional[int] = 12
    checksum_cache_ttl: Optional[int] = 60 * 60 * 24 * 1 # 1 days
    enable_progress_bar: Optional[bool] = False
    tfio_enabled: Optional[bool] = False

    @lazyproperty
    def core(self) -> CoreSettings:
        return CoreSettings()
    
    @lazyproperty
    def aws(self) -> AwsSettings:
        return AwsSettings()

    @lazyproperty
    def gcp(self) -> GcpSettings:
        return GcpSettings()

    @lazyproperty
    def azure(self) -> AzureSettings:
        return AzureSettings()

    @lazyproperty
    def minio(self) -> MinioSettings:
        return MinioSettings()
    
    @lazyproperty
    def s3_compat(self) -> S3CompatSettings:
        return S3CompatSettings()
    
    @lazyproperty
    def r2(self) -> CloudflareR2Settings:
        return CloudflareR2Settings()
    
    @lazyproperty
    def wasabi(self) -> WasabiS3Settings:
        return WasabiS3Settings()
    
    @lazyproperty
    def github(self) -> GithubSettings:
        return GithubSettings()
    
    @lazyproperty
    def huggingface(self) -> HuggingfaceSettings:
        return HuggingfaceSettings()

    def create_adc(
        self, 
        data: Union[str, Dict[str, Any]], 
        path: str = None
    ):
        """
        Create a new ADC based on the passed data and writes it to 
        path or GOOGLE_APPLICATION_CREDENTIALS
        """
        if isinstance(data, str): data = json.loads(data)
        path: pathlib.Path = pathlib.Path(path) if path else self.gcp.google_application_credentials
        path.write_text(json.dumps(data, indent = 2, ensure_ascii=False))
        self.gcp.google_application_credentials = path

    def get_boto_values(self):
        t = "[Credentials]\n"
        if self.aws.aws_access_key_id:
            t += f"aws_access_key_id = {self.aws.aws_access_key_id}\n"
        if self.aws.aws_secret_access_key:
            t += f"aws_secret_access_key = {self.aws.aws_secret_access_key}\n"
        if self.gcp.google_application_credentials.exists():
            t += f"gs_service_key_file = {self.gcp.google_application_credentials.as_posix()}\n"
        t += "\n[Boto]\n"
        t += "https_validate_certificates = True\n"
        t += "\n[GSUtil]\n"
        t += "content_language = en\n"
        t += "default_api_version = 2\n"
        if self.gcp.project: t += f"default_project_id = {self.gcp.project}\n"
        return t

    def write_botofile(
        self, 
        overwrite: bool = False, 
        **kwargs
    ):
        if not self.core.boto_config_exists or overwrite:
            logger.info(f"Writing boto config to {self.core.boto_config_path.as_posix()}")
            self.core.boto_config_path.write_text(self.get_boto_values())

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            elif isinstance(getattr(self, k), BaseSettings):
                getattr(self, k).update_config(**v)
            else: setattr(self, k, v)

    def set_env(self):
        self.aws.set_env()
        self.gcp.set_env()
        self.minio.set_env()
        self.s3_compat.set_env()
        # self.github.set_env()
        # self.huggingface.set_env()
        self.r2.set_env()
        self.wasabi.set_env()

    def update_auth(self, update_fs: bool = True, **config):
        self.update_config(**config)
        self.set_env()

        if update_fs:
            # Reset the accessor to use the new settings
            from fileio.lib.posix.filesys import FileSysManager
            # from fileio.providers.filesys import get_accessor
            if config.get('aws'):
                FileSysManager.get_accessor('s3', _reset = True)
                # get_accessor('s3', _reset = True)
            if config.get('gcp'):
                FileSysManager.get_accessor('gs', _reset = True)
                # get_accessor('gs', _reset = True)
            if config.get('minio'):
                FileSysManager.get_accessor('minio', _reset = True)
                # get_accessor('minio', _reset = True)
            
            if config.get('s3_compat') or config.get('s3c'):
                FileSysManager.get_accessor('s3c', _reset = True)
                # get_accessor('s3_compat', _reset = True)
            
            if config.get('r2'):
                FileSysManager.get_accessor('r2', _reset = True)
                # get_accessor('r2', _reset = True)

            if config.get('wasabi'):
                FileSysManager.get_accessor('wasabi', _reset = True)
    

    class Config(BaseSettings.Config):
        env_prefix = "FILEIO_"
        case_sensitive = False
    

settings = Settings()


from __future__ import annotations

"""
FileIO CLI
"""

import typer
import asyncio

from typing import Optional

_help = """
FileIO CLI

Usage:
    fileio <command> [options]

Commands:
    copy: Copy a file
    info: Get information about a file or directory
    ls: List files in a directory
    rm: Remove a file or directory
    cat: Echo the file contents

"""

cmd = typer.Typer(no_args_is_help = True, help = _help)

@cmd.command('copy', help = 'Copy a file')
def copy_file(
    source: str = typer.Argument(..., help = 'Path to file to copy'),
    destination: str = typer.Argument(..., help = 'Path to destination file. Can be a directory or a file'),
    filename: Optional[str] = typer.Option(None, help = 'Name of file to upload as. If not provided, will use the name of the source file'),
    new_suffix: Optional[str] = typer.Option(None, help = 'If provided, will replace the suffix of the source file with this suffix'),
    overwrite: Optional[bool] = typer.Option(False, help = 'Whether to overwrite the destination file if it exists'),
):
    """
    Usage:
    Upload or Download a file using FileIO backend

    $ fileio get <source> <destination> [options]

    """
    from .functions import fetch_file
    asyncio.run(
        fetch_file(
            source = source,
            destination = destination,
            filename = filename,
            new_suffix = new_suffix,
            overwrite = overwrite,
        )
    )


@cmd.command('info', help = 'Get information about a file or directory')
def source_info_cmd(
    source: str = typer.Argument(..., help = 'Path to file or directory'),
    recursive: Optional[bool] = typer.Option(False, help = 'Whether to recursively list files'),
    dirs_only: Optional[bool] = typer.Option(False, help = 'Whether to only dirs (default is files only)'),
):
    """
    Usage:
    Get information about a file or directory using FileIO backend

    $ fileio info <source> [options]

    """
    from .functions import source_info
    asyncio.run(
        source_info(
            source = source,
            recursive = recursive,
            dirs_only = dirs_only,
        )
    )


@cmd.command('ls', help = 'List files in a directory')
def list_files(
    source: str = typer.Argument(..., help = 'Path to directory'),
    pattern: str = typer.Argument('*', help = 'Pattern to match files against'),
):
    """
    Usage:
    List files in a directory using FileIO backend

    $ fileio ls <source> [options]

    """
    from .functions import list_dir
    asyncio.run(
        list_dir(
            source = source,
            pattern = pattern,
        )
    )

@cmd.command('rm', help = 'Remove a file or directory')
def rm_cmd(
    source: str = typer.Argument(..., help = 'Path to file or directory'),
    recursive: Optional[bool] = typer.Option(False, help = 'Whether to recursively delete files'),
    force: Optional[bool] = typer.Option(False, help = 'Whether to force delete without prompting'),
):
    """
    Usage:
    Delete a file or directory using FileIO backend

    $ fileio rm <source> [options]

    """
    from .functions import rm_files_or_dir
    asyncio.run(
        rm_files_or_dir(
            source = source,
            recursive = recursive,
            force = force,
        )
    )


@cmd.command('cat', help = 'Echo the file contents')
def cat_cmd(
    source: str = typer.Argument(..., help = 'Path to file'),
    binary: Optional[bool] = typer.Option(False, help = 'Whether to read the file as binary'),
    encode: Optional[str] = typer.Option(None, help = 'Encoding to use when reading the file'),
    decode: Optional[str] = typer.Option(None, help = 'Decoding to use when writing the file'),
):
    """
    Usage:
    Echo the file contents using FileIO backend

    $ fileio cat <source> [options]

    """
    from .functions import cat_file
    asyncio.run(
        cat_file(
            source = source,
            binary = binary,
            encode = encode,
            decode = decode,
        )
    )


if __name__ == '__main__':
    cmd()







# handle posix stat stuff
import os
import sys
#import stat
import time
import inspect
import random
import typing
import asyncio
import hashlib
import functools
import contextlib
import datetime
#from datetime import datetime
#from urllib.parse import quote
from email.utils import formatdate
from mimetypes import guess_type as mimetypes_guess_type

from frozendict import frozendict
from fileio.utils.logs import default_logger


try:
    # check if the Python version supports the parameter
    # using usedforsecurity=False to avoid an exception on FIPS systems
    # that reject usedforsecurity=True
    hashlib.md5(b"data", usedforsecurity=False)  # type: ignore[call-arg]

    def md5_hexdigest(
        data: bytes, *, usedforsecurity: bool = True
    ) -> str:  # pragma: no cover
        return hashlib.md5(  # type: ignore[call-arg]
            data, usedforsecurity=usedforsecurity
        ).hexdigest()

except TypeError:  # pragma: no cover

    def md5_hexdigest(data: bytes, *, usedforsecurity: bool = True) -> str:
        return hashlib.md5(data).hexdigest()

# Compatibility wrapper for `mimetypes.guess_type` to support `os.PathLike` on <py3.8
def guess_type(
    path: typing.Union[str, "os.PathLike[str]"], strict: bool = True
) -> typing.Tuple[typing.Optional[str], typing.Optional[str]]:
    if sys.version_info < (3, 8):  # pragma: no cover
        path = os.fspath(path)
    return mimetypes_guess_type(path, strict)


def get_file_info(path: typing.Union[str, "os.PathLike[str]"]) -> typing.Dict[str, typing.Any]:
    if sys.version_info < (3, 8):  # pragma: no cover
        path = os.fspath(path)
    stat_result = os.stat(path)
    last_modified = formatdate(stat_result.st_mtime, usegmt=True)
    etag_base = f"{str(stat_result.st_mtime)}-{str(stat_result.st_size)}"
    etag = md5_hexdigest(etag_base.encode(), usedforsecurity=False)
    return {
        "size": stat_result.st_size,
        "last-modified": last_modified,
        "etag": etag,
        "media-type": guess_type(path)[0] or "text/plain",
        "filename": os.path.basename(path),
    }

def timer(t: typing.Optional[float] = None, msg: typing.Optional[str] = None, logger = default_logger):
    if not t: return time.perf_counter()
    done_time = time.perf_counter() - t
    if msg: logger.info(f'{msg} in {done_time:.2f} secs')
    return done_time

def timed(func: typing.Callable):
    """
    Decorator to time a function
    """
    _func_name = func.__name__
    @functools.wraps(func)
    async def fx(*args, **kwargs):
        start = time.perf_counter()
        if inspect.iscoroutinefunction(func): result = await func(*args, **kwargs)
        else: result = func(*args, **kwargs)
        end = time.perf_counter()
        default_logger.info(f'{_func_name}: {end - start:.4f} secs')
        return result
    return fx


def exponential_backoff(
    attempts: int,
    base_delay: int = 1,
    max_delay: int = None,
    jitter: bool = True,
):
    """
    Get the next delay for retries in exponential backoff.

    attempts: Number of attempts so far
    base_delay: Base delay, in seconds
    max_delay: Max delay, in seconds. If None (default), there is no max.
    jitter: If True, add a random jitter to the delay
    """
    if max_delay is None:
        max_delay = float("inf")
    backoff = min(max_delay, base_delay * 2 ** max(attempts - 1, 0))
    if jitter:
        backoff = backoff * random.random()
    return backoff

def retryable(limit: int = 3, delay: int = 3):
    def decorator(func: typing.Callable):
        if not inspect.iscoroutinefunction(func):
            def sync_wrapper(*args, **kwargs):
                for n in range(limit - 1):
                    with contextlib.suppress(Exception):
                        return func(*args, **kwargs)
                    time.sleep(exponential_backoff(n, base_delay=delay))
                return func(*args, **kwargs)
            return sync_wrapper
        else:
            async def async_wrapper(*args, **kwargs):
                for n in range(limit-1):
                    with contextlib.suppress(Exception):
                        return await func(*args, **kwargs)
                    await asyncio.sleep(exponential_backoff(n, base_delay=delay))
                return await func(*args, **kwargs)
            return async_wrapper
    return decorator


def create_timestamp(
    tz: typing.Optional[datetime.tzinfo] = datetime.timezone.utc,
    as_str: typing.Optional[bool] = False,
):
    """
    Creates a timestamp
    args:
        tz: timezone
        as_str: if True, returns a string
    """
    dt = datetime.datetime.now(tz =tz)
    return dt.isoformat() if as_str else dt

def is_coro_func(obj, func_name: str = None):
    """
    This is probably in the library elsewhere but returns bool
    based on if the function is a coro
    """
    try:
        if inspect.iscoroutinefunction(obj): return True
        if inspect.isawaitable(obj): return True
        if func_name and hasattr(obj, func_name) and inspect.iscoroutinefunction(getattr(obj, func_name)):
            return True
        return bool(hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__))

    except Exception:
        return False


"""
Timed Caches
"""

def recursive_freeze(value):
    if not isinstance(value, dict):
        return value
    for k,v in value.items():
        value[k] = recursive_freeze(v)
    return frozendict(value)

# To unfreeze
def recursive_unfreeze(value):
    if isinstance(value, frozendict):
        value = dict(value)
        for k,v in value.items():
            value[k] = recursive_unfreeze(v)
    
    return value

def freeze_args_and_kwargs(*args, **kwargs):
    args = tuple(
        recursive_freeze(arg) if isinstance(arg, dict) else arg
        for arg in args
    )
    kwargs = {k: recursive_freeze(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
    return args, kwargs



def timed_cache(
    secs: typing.Optional[int] = 60 * 60, 
    maxsize: int = 1024
):
    """
    Wrapper for creating a expiring cached function
    args:
        secs: number of seconds to cache the function
        maxsize: maxsize of the cache
    """
    if secs is None: secs = 60 * 60
    def wrapper_cache(func):
        func = functools.lru_cache(maxsize=maxsize)(func)
        func.lifetime = datetime.timedelta(seconds=secs)
        func.expiration = create_timestamp() + func.lifetime
        def _check_cache(func):
            if create_timestamp() >= func.expiration:
                func.cache_clear()
                func.expiration = create_timestamp() + func.lifetime

        if is_coro_func(func):
            # https://stackoverflow.com/questions/34116942/how-to-cache-asyncio-coroutines
            @functools.wraps(func)
            def wrapped_func(*args, **kwargs):
                _check_cache(func)
                args, kwargs = freeze_args_and_kwargs(*args, **kwargs)
                coro = func(*args, **kwargs)
                return asyncio.ensure_future(coro)
                # return await func(*args, **kwargs)
            return wrapped_func

        else:
            @functools.wraps(func)
            def wrapped_func(*args, **kwargs):
                _check_cache(func)
                args, kwargs = freeze_args_and_kwargs(*args, **kwargs)
                return func(*args, **kwargs)
            return wrapped_func

    return wrapper_cache


__all__ = [
    "get_file_info",
    "md5_hexdigest",
    "mimetypes_guess_type",
    "guess_type",
    'timer',
    'timed',
    'exponential_backoff',
    'retryable',
    'create_timestamp',

]



# https://en.wikipedia.org/wiki/Ligature_(writing)
KNOWN_LIGATURES = {
    # Latin
    "ﬀ": "ff",
    "ﬁ": "fi",
    "ﬂ": "fl",
    "ﬃ": "ffi",
    "ﬄ": "ffl",
    "ﬅ": "ft",
    "ﬆ": "st",
    "Ǳ": "DZ",
    "ǲ": "Dz",
    "ǳ": "dz",
    "Ǆ": "DŽ",
    "ǅ": "Dž",
    "ǆ": "dž",
    "Ꜩ": "Tz",
    "ꜩ": "tz",
    "🙰": "et",
    "℔": "lb",
    "ᵫ": "ue",
    "Ĳ": "IJ",
    "ĳ": "ij",  # They are both capitalized together, so the "Ij" ligature doesn't exist
    "ꝏ": "oo",  # Not the infinite sign but a double-o ligature: https://en.wikipedia.org/wiki/Ligature_(writing)#Massachusett_%EA%9D%8F
    # Armenian
    "ﬓ": "մն",
    "ﬔ": "մե",
    "ﬕ": "մի",
    "ﬖ": "վն",
    "ﬗ": "մխ",
}



import base64
import gzip

from typing import Union, Optional
from .base import BasePack


class Base64GZip(BasePack):
    encoding: str = 'utf-8'

    @classmethod
    def encode(cls, data: Union[str, bytes], encoding: str = 'utf-8', *args, as_bytes: Optional[bool] = False, **kwargs) -> Union[str, bytes]:
        if isinstance(data, str): data = data.encode(encoding = encoding)
        data = base64.b64encode(gzip.compress(data, *args, **kwargs))
        return data if as_bytes else data.decode(encoding = encoding)

    @classmethod
    def decode(cls, data: Union[str, bytes], encoding: str = 'utf-8', *args, as_bytes: Optional[bool] = False, **kwargs) -> Union[str, bytes]:
        if isinstance(data, str): data = data.encode(encoding = encoding)
        data = gzip.decompress(base64.b64decode(data, *args, **kwargs))
        return data if as_bytes else data.decode(encoding = encoding)

    @classmethod
    def dumps(cls, data: Union[str, bytes], encoding: str = 'utf-8', *args, as_bytes: Optional[bool] = False, **kwargs) -> Union[str, bytes]:
        return cls.encode(data, encoding = encoding, *args, as_bytes = as_bytes, **kwargs)
    
    @classmethod
    def loads(cls, data: Union[str, bytes], encoding: str = 'utf-8', *args, as_bytes: Optional[bool] = False, **kwargs) -> Union[str, bytes]:
        return cls.decode(data, encoding = encoding, *args, as_bytes = as_bytes, **kwargs)

